"""
Copyright (C) Cross The Road Electronics.  All rights reserved.
License information can be found in CTRE_LICENSE.txt
For support and suggestions contact support@ctr-electronics.com or file
an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
"""

from phoenix6.hardware.parent_device import ParentDevice, SupportsSendRequest
from phoenix6.spns.spn_value import SpnValue
from phoenix6.status_code import StatusCode
from phoenix6.status_signal import *
from phoenix6.units import *
from phoenix6.sim.device_type import DeviceType
from phoenix6.configs.pigeon2_configs import Pigeon2Configurator
from phoenix6.sim.pigeon2_sim_state import Pigeon2SimState

class CorePigeon2(ParentDevice):
    """
    Constructs a new Pigeon 2 sensor object.

    :param device_id: ID of the device, as configured in Phoenix Tuner.
    :type device_id: int
    :param canbus: Name of the CAN bus this device is on. Possible CAN bus strings are:
                       rio - The native roboRIO CAN bus
                       CANivore Name or Serial Number
                       SocketCAN interface - non-FRC Linux only
                       * - Any CANivore seen by the program
                       Empty String - Default for the system ("rio" for roboRIO, "can0" for linux, "*" for Windows)
    :type canbus: str, optional
    """

    def __init__(self, device_id: int, canbus: str = ""):
        super().__init__(device_id, "pigeon 2", canbus)
        self.configurator = Pigeon2Configurator(self._device_identifier)

        Native.instance().c_ctre_phoenix6_platform_sim_create(DeviceType.PRO_Pigeon2Type.value, device_id)
        self.__sim_state = None


    @property
    def sim_state(self) -> Pigeon2SimState:
        """
        Get the simulation state for this device.

        This function reuses an allocated simulation state
        object, so it is safe to call this function multiple
        times in a robot loop.

        :returns: Simulation state
        :rtype: Pigeon2SimState
        """

        if self.__sim_state is None:
            self.__sim_state = Pigeon2SimState(self)
        return self.__sim_state


    

    def get_version_major(self) -> StatusSignal[int]:
        """
        App Major Version number.
        
          Minimum Value: 0
          Maximum Value: 255
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: VersionMajor Status Signal Object
        :rtype: StatusSignal[int]
        """
        return self._common_lookup(SpnValue.VERSION_MAJOR.value, 0, None, "version_major", False, int)
    
    def get_version_minor(self) -> StatusSignal[int]:
        """
        App Minor Version number.
        
          Minimum Value: 0
          Maximum Value: 255
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: VersionMinor Status Signal Object
        :rtype: StatusSignal[int]
        """
        return self._common_lookup(SpnValue.VERSION_MINOR.value, 0, None, "version_minor", False, int)
    
    def get_version_bugfix(self) -> StatusSignal[int]:
        """
        App Bugfix Version number.
        
          Minimum Value: 0
          Maximum Value: 255
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: VersionBugfix Status Signal Object
        :rtype: StatusSignal[int]
        """
        return self._common_lookup(SpnValue.VERSION_BUGFIX.value, 0, None, "version_bugfix", False, int)
    
    def get_version_build(self) -> StatusSignal[int]:
        """
        App Build Version number.
        
          Minimum Value: 0
          Maximum Value: 255
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: VersionBuild Status Signal Object
        :rtype: StatusSignal[int]
        """
        return self._common_lookup(SpnValue.VERSION_BUILD.value, 0, None, "version_build", False, int)
    
    def get_version(self) -> StatusSignal[int]:
        """
        Full Version.  The format is a four byte value.
        
        Full Version of firmware in device. The format is a four byte value.
        
          Minimum Value: 0
          Maximum Value: 4294967295
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Version Status Signal Object
        :rtype: StatusSignal[int]
        """
        return self._common_lookup(SpnValue.VERSION_FULL.value, 0, None, "version", False, int)
    
    def get_fault_field(self) -> StatusSignal[int]:
        """
        Integer representing all faults
        
        This returns the fault flags reported by the device. These are device
        specific and are not used directly in typical applications. Use the
        signal specific GetFault_*() methods instead.  
        
          Minimum Value: 0
          Maximum Value: 16777215
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: FaultField Status Signal Object
        :rtype: StatusSignal[int]
        """
        return self._common_lookup(SpnValue.ALL_FAULTS.value, 0, None, "fault_field", True, int)
    
    def get_sticky_fault_field(self) -> StatusSignal[int]:
        """
        Integer representing all sticky faults
        
        This returns the persistent "sticky" fault flags reported by the
        device. These are device specific and are not used directly in typical
        applications. Use the signal specific GetStickyFault_*() methods
        instead.  
        
          Minimum Value: 0
          Maximum Value: 16777215
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFaultField Status Signal Object
        :rtype: StatusSignal[int]
        """
        return self._common_lookup(SpnValue.ALL_STICKY_FAULTS.value, 0, None, "sticky_fault_field", True, int)
    
    def get_yaw(self) -> StatusSignal[degree]:
        """
        Current reported yaw of the Pigeon2.
        
          Minimum Value: -368640.0
          Maximum Value: 368639.99725341797
          Default Value: 0
          Units: deg
        
        Default Rates:
          CAN 2.0: 100.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: Yaw Status Signal Object
        :rtype: StatusSignal[degree]
        """
        return self._common_lookup(SpnValue.PIGEON2_YAW.value, 0, None, "yaw", True, degree)
    
    def get_pitch(self) -> StatusSignal[degree]:
        """
        Current reported pitch of the Pigeon2.
        
          Minimum Value: -90.0
          Maximum Value: 89.9560546875
          Default Value: 0
          Units: deg
        
        Default Rates:
          CAN 2.0: 100.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: Pitch Status Signal Object
        :rtype: StatusSignal[degree]
        """
        return self._common_lookup(SpnValue.PIGEON2_PITCH.value, 0, None, "pitch", True, degree)
    
    def get_roll(self) -> StatusSignal[degree]:
        """
        Current reported roll of the Pigeon2.
        
          Minimum Value: -180.0
          Maximum Value: 179.9560546875
          Default Value: 0
          Units: deg
        
        Default Rates:
          CAN 2.0: 100.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: Roll Status Signal Object
        :rtype: StatusSignal[degree]
        """
        return self._common_lookup(SpnValue.PIGEON2_ROLL.value, 0, None, "roll", True, degree)
    
    def get_quat_w(self) -> StatusSignal[float]:
        """
        The W component of the reported Quaternion.
        
          Minimum Value: -1.0001220852154804
          Maximum Value: 1.0
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN 2.0: 50.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: QuatW Status Signal Object
        :rtype: StatusSignal[float]
        """
        return self._common_lookup(SpnValue.PIGEON2_QUATW.value, 0, None, "quat_w", True, float)
    
    def get_quat_x(self) -> StatusSignal[float]:
        """
        The X component of the reported Quaternion.
        
          Minimum Value: -1.0001220852154804
          Maximum Value: 1.0
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN 2.0: 50.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: QuatX Status Signal Object
        :rtype: StatusSignal[float]
        """
        return self._common_lookup(SpnValue.PIGEON2_QUATX.value, 0, None, "quat_x", True, float)
    
    def get_quat_y(self) -> StatusSignal[float]:
        """
        The Y component of the reported Quaternion.
        
          Minimum Value: -1.0001220852154804
          Maximum Value: 1.0
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN 2.0: 50.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: QuatY Status Signal Object
        :rtype: StatusSignal[float]
        """
        return self._common_lookup(SpnValue.PIGEON2_QUATY.value, 0, None, "quat_y", True, float)
    
    def get_quat_z(self) -> StatusSignal[float]:
        """
        The Z component of the reported Quaternion.
        
          Minimum Value: -1.0001220852154804
          Maximum Value: 1.0
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN 2.0: 50.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: QuatZ Status Signal Object
        :rtype: StatusSignal[float]
        """
        return self._common_lookup(SpnValue.PIGEON2_QUATZ.value, 0, None, "quat_z", True, float)
    
    def get_gravity_vector_x(self) -> StatusSignal[float]:
        """
        The X component of the gravity vector.
        
        This is the X component of the reported gravity-vector. The gravity
        vector is not the acceleration experienced by the Pigeon2, rather it
        is where the Pigeon2 believes "Down" is. This can be used for
        mechanisms that are linearly related to gravity, such as an arm
        pivoting about a point, as the contribution of gravity to the arm is
        directly proportional to the contribution of gravity about one of
        these primary axis.
        
          Minimum Value: -1.000030518509476
          Maximum Value: 1.0
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: GravityVectorX Status Signal Object
        :rtype: StatusSignal[float]
        """
        return self._common_lookup(SpnValue.PIGEON2_GRAVITY_VECTORX.value, 0, None, "gravity_vector_x", True, float)
    
    def get_gravity_vector_y(self) -> StatusSignal[float]:
        """
        The Y component of the gravity vector.
        
        This is the X component of the reported gravity-vector. The gravity
        vector is not the acceleration experienced by the Pigeon2, rather it
        is where the Pigeon2 believes "Down" is. This can be used for
        mechanisms that are linearly related to gravity, such as an arm
        pivoting about a point, as the contribution of gravity to the arm is
        directly proportional to the contribution of gravity about one of
        these primary axis.
        
          Minimum Value: -1.000030518509476
          Maximum Value: 1.0
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: GravityVectorY Status Signal Object
        :rtype: StatusSignal[float]
        """
        return self._common_lookup(SpnValue.PIGEON2_GRAVITY_VECTORY.value, 0, None, "gravity_vector_y", True, float)
    
    def get_gravity_vector_z(self) -> StatusSignal[float]:
        """
        The Z component of the gravity vector.
        
        This is the Z component of the reported gravity-vector. The gravity
        vector is not the acceleration experienced by the Pigeon2, rather it
        is where the Pigeon2 believes "Down" is. This can be used for
        mechanisms that are linearly related to gravity, such as an arm
        pivoting about a point, as the contribution of gravity to the arm is
        directly proportional to the contribution of gravity about one of
        these primary axis.
        
          Minimum Value: -1.000030518509476
          Maximum Value: 1.0
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: GravityVectorZ Status Signal Object
        :rtype: StatusSignal[float]
        """
        return self._common_lookup(SpnValue.PIGEON2_GRAVITY_VECTORZ.value, 0, None, "gravity_vector_z", True, float)
    
    def get_temperature(self) -> StatusSignal[celsius]:
        """
        Temperature of the Pigeon 2.
        
          Minimum Value: -128.0
          Maximum Value: 127.99609375
          Default Value: 0
          Units: ℃
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: Temperature Status Signal Object
        :rtype: StatusSignal[celsius]
        """
        return self._common_lookup(SpnValue.PIGEON2_TEMPERATURE.value, 0, None, "temperature", True, celsius)
    
    def get_no_motion_enabled(self) -> StatusSignal[bool]:
        """
        Whether the no-motion calibration feature is enabled.
        
          Default Value: 0
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: NoMotionEnabled Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.PIGEON2_NO_MOTION_CAL_ENABLED.value, 0, None, "no_motion_enabled", True, bool)
    
    def get_no_motion_count(self) -> StatusSignal[float]:
        """
        The number of times a no-motion event occurred, wraps at 15.
        
          Minimum Value: 0
          Maximum Value: 15
          Default Value: 0
          Units: 
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: NoMotionCount Status Signal Object
        :rtype: StatusSignal[float]
        """
        return self._common_lookup(SpnValue.PIGEON2_NO_MOTION_COUNT.value, 0, None, "no_motion_count", True, float)
    
    def get_temperature_compensation_disabled(self) -> StatusSignal[bool]:
        """
        Whether the temperature-compensation feature is disabled.
        
          Default Value: 0
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: TemperatureCompensationDisabled Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.PIGEON2_TEMP_COMP_DISABLED.value, 0, None, "temperature_compensation_disabled", True, bool)
    
    def get_up_time(self) -> StatusSignal[second]:
        """
        How long the Pigeon 2's been up in seconds, caps at 255 seconds.
        
          Minimum Value: 0
          Maximum Value: 255
          Default Value: 0
          Units: s
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: UpTime Status Signal Object
        :rtype: StatusSignal[second]
        """
        return self._common_lookup(SpnValue.PIGEON2_UP_TIME.value, 0, None, "up_time", True, second)
    
    def get_accum_gyro_x(self) -> StatusSignal[degree]:
        """
        The accumulated gyro about the X axis without any sensor fusing.
        
          Minimum Value: -23040.0
          Maximum Value: 23039.9560546875
          Default Value: 0
          Units: deg
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AccumGyroX Status Signal Object
        :rtype: StatusSignal[degree]
        """
        return self._common_lookup(SpnValue.PIGEON2_ACCUM_GYROX.value, 0, None, "accum_gyro_x", True, degree)
    
    def get_accum_gyro_y(self) -> StatusSignal[degree]:
        """
        The accumulated gyro about the Y axis without any sensor fusing.
        
          Minimum Value: -23040.0
          Maximum Value: 23039.9560546875
          Default Value: 0
          Units: deg
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AccumGyroY Status Signal Object
        :rtype: StatusSignal[degree]
        """
        return self._common_lookup(SpnValue.PIGEON2_ACCUM_GYROY.value, 0, None, "accum_gyro_y", True, degree)
    
    def get_accum_gyro_z(self) -> StatusSignal[degree]:
        """
        The accumulated gyro about the Z axis without any sensor fusing.
        
          Minimum Value: -23040.0
          Maximum Value: 23039.9560546875
          Default Value: 0
          Units: deg
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AccumGyroZ Status Signal Object
        :rtype: StatusSignal[degree]
        """
        return self._common_lookup(SpnValue.PIGEON2_ACCUM_GYROZ.value, 0, None, "accum_gyro_z", True, degree)
    
    def get_angular_velocity_x_world(self) -> StatusSignal[degrees_per_second]:
        """
        Angular Velocity world X
        
        This is the X component of the angular velocity with respect to the
        world frame and is mount-calibrated.
        
          Minimum Value: -2048.0
          Maximum Value: 2047.99609375
          Default Value: 0
          Units: dps
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AngularVelocityXWorld Status Signal Object
        :rtype: StatusSignal[degrees_per_second]
        """
        return self._common_lookup(SpnValue.PIGEON2_ANGULAR_VELOCITY_XWORLD.value, 0, None, "angular_velocity_x_world", True, degrees_per_second)
    
    def get_angular_velocity_y_world(self) -> StatusSignal[degrees_per_second]:
        """
        Angular Velocity Quaternion Y Component
        
        This is the Y component of the angular velocity with respect to the
        world frame and is mount-calibrated.
        
          Minimum Value: -2048.0
          Maximum Value: 2047.99609375
          Default Value: 0
          Units: dps
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AngularVelocityYWorld Status Signal Object
        :rtype: StatusSignal[degrees_per_second]
        """
        return self._common_lookup(SpnValue.PIGEON2_ANGULAR_VELOCITY_YWORLD.value, 0, None, "angular_velocity_y_world", True, degrees_per_second)
    
    def get_angular_velocity_z_world(self) -> StatusSignal[degrees_per_second]:
        """
        Angular Velocity Quaternion Z Component
        
        This is the Z component of the angular velocity with respect to the
        world frame and is mount-calibrated.
        
          Minimum Value: -2048.0
          Maximum Value: 2047.99609375
          Default Value: 0
          Units: dps
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AngularVelocityZWorld Status Signal Object
        :rtype: StatusSignal[degrees_per_second]
        """
        return self._common_lookup(SpnValue.PIGEON2_ANGULAR_VELOCITY_ZWORLD.value, 0, None, "angular_velocity_z_world", True, degrees_per_second)
    
    def get_acceleration_x(self) -> StatusSignal[g]:
        """
        The acceleration measured by Pigeon2 in the X direction.
        
        This value includes the acceleration due to gravity. If this is
        undesirable, get the gravity vector and subtract out the contribution
        in this direction.
        
          Minimum Value: -2.0
          Maximum Value: 1.99993896484375
          Default Value: 0
          Units: g
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AccelerationX Status Signal Object
        :rtype: StatusSignal[g]
        """
        return self._common_lookup(SpnValue.PIGEON2_ACCELERATIONX.value, 0, None, "acceleration_x", True, g)
    
    def get_acceleration_y(self) -> StatusSignal[g]:
        """
        The acceleration measured by Pigeon2 in the Y direction.
        
        This value includes the acceleration due to gravity. If this is
        undesirable, get the gravity vector and subtract out the contribution
        in this direction.
        
          Minimum Value: -2.0
          Maximum Value: 1.99993896484375
          Default Value: 0
          Units: g
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AccelerationY Status Signal Object
        :rtype: StatusSignal[g]
        """
        return self._common_lookup(SpnValue.PIGEON2_ACCELERATIONY.value, 0, None, "acceleration_y", True, g)
    
    def get_acceleration_z(self) -> StatusSignal[g]:
        """
        The acceleration measured by Pigeon2 in the Z direction.
        
        This value includes the acceleration due to gravity. If this is
        undesirable, get the gravity vector and subtract out the contribution
        in this direction.
        
          Minimum Value: -2.0
          Maximum Value: 1.99993896484375
          Default Value: 0
          Units: g
        
        Default Rates:
          CAN 2.0: 10.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: AccelerationZ Status Signal Object
        :rtype: StatusSignal[g]
        """
        return self._common_lookup(SpnValue.PIGEON2_ACCELERATIONZ.value, 0, None, "acceleration_z", True, g)
    
    def get_supply_voltage(self) -> StatusSignal[volt]:
        """
        Measured supply voltage to the Pigeon2.
        
          Minimum Value: 0.0
          Maximum Value: 31.99951171875
          Default Value: 0
          Units: V
        
        Default Rates:
          CAN 2.0: 4.0 Hz
          CAN FD: 100.0 Hz (TimeSynced with Pro)
        
        :returns: SupplyVoltage Status Signal Object
        :rtype: StatusSignal[volt]
        """
        return self._common_lookup(SpnValue.PIGEON2_SUPPLY_VOLTAGE.value, 0, None, "supply_voltage", True, volt)
    
    def get_angular_velocity_x_device(self) -> StatusSignal[degrees_per_second]:
        """
        The angular velocity (ω) of the Pigeon 2 about the device's X axis.
        
        This value is not mount-calibrated
        
          Minimum Value: -1998.048780487805
          Maximum Value: 1997.987804878049
          Default Value: 0
          Units: dps
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: AngularVelocityXDevice Status Signal Object
        :rtype: StatusSignal[degrees_per_second]
        """
        return self._common_lookup(SpnValue.PIGEON2_ANGULAR_VELOCITYX.value, 0, None, "angular_velocity_x_device", True, degrees_per_second)
    
    def get_angular_velocity_y_device(self) -> StatusSignal[degrees_per_second]:
        """
        The angular velocity (ω) of the Pigeon 2 about the device's Y axis.
        
        This value is not mount-calibrated
        
          Minimum Value: -1998.048780487805
          Maximum Value: 1997.987804878049
          Default Value: 0
          Units: dps
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: AngularVelocityYDevice Status Signal Object
        :rtype: StatusSignal[degrees_per_second]
        """
        return self._common_lookup(SpnValue.PIGEON2_ANGULAR_VELOCITYY.value, 0, None, "angular_velocity_y_device", True, degrees_per_second)
    
    def get_angular_velocity_z_device(self) -> StatusSignal[degrees_per_second]:
        """
        The angular velocity (ω) of the Pigeon 2 about the device's Z axis.
        
        This value is not mount-calibrated
        
          Minimum Value: -1998.048780487805
          Maximum Value: 1997.987804878049
          Default Value: 0
          Units: dps
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: AngularVelocityZDevice Status Signal Object
        :rtype: StatusSignal[degrees_per_second]
        """
        return self._common_lookup(SpnValue.PIGEON2_ANGULAR_VELOCITYZ.value, 0, None, "angular_velocity_z_device", True, degrees_per_second)
    
    def get_magnetic_field_x(self) -> StatusSignal[microtesla]:
        """
        The biased magnitude of the magnetic field measured by the Pigeon 2 in
        the X direction. This is only valid after performing a magnetometer
        calibration.
        
          Minimum Value: -19660.8
          Maximum Value: 19660.2
          Default Value: 0
          Units: uT
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: MagneticFieldX Status Signal Object
        :rtype: StatusSignal[microtesla]
        """
        return self._common_lookup(SpnValue.PIGEON2_MAGNETIC_FIELDX.value, 0, None, "magnetic_field_x", True, microtesla)
    
    def get_magnetic_field_y(self) -> StatusSignal[microtesla]:
        """
        The biased magnitude of the magnetic field measured by the Pigeon 2 in
        the Y direction. This is only valid after performing a magnetometer
        calibration.
        
          Minimum Value: -19660.8
          Maximum Value: 19660.2
          Default Value: 0
          Units: uT
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: MagneticFieldY Status Signal Object
        :rtype: StatusSignal[microtesla]
        """
        return self._common_lookup(SpnValue.PIGEON2_MAGNETIC_FIELDY.value, 0, None, "magnetic_field_y", True, microtesla)
    
    def get_magnetic_field_z(self) -> StatusSignal[microtesla]:
        """
        The biased magnitude of the magnetic field measured by the Pigeon 2 in
        the Z direction. This is only valid after performing a magnetometer
        calibration.
        
          Minimum Value: -19660.8
          Maximum Value: 19660.2
          Default Value: 0
          Units: uT
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: MagneticFieldZ Status Signal Object
        :rtype: StatusSignal[microtesla]
        """
        return self._common_lookup(SpnValue.PIGEON2_MAGNETIC_FIELDZ.value, 0, None, "magnetic_field_z", True, microtesla)
    
    def get_raw_magnetic_field_x(self) -> StatusSignal[microtesla]:
        """
        The raw magnitude of the magnetic field measured by the Pigeon 2 in
        the X direction. This is only valid after performing a magnetometer
        calibration.
        
          Minimum Value: -19660.8
          Maximum Value: 19660.2
          Default Value: 0
          Units: uT
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: RawMagneticFieldX Status Signal Object
        :rtype: StatusSignal[microtesla]
        """
        return self._common_lookup(SpnValue.PIGEON2_RAW_MAGNETIC_FIELDX.value, 0, None, "raw_magnetic_field_x", True, microtesla)
    
    def get_raw_magnetic_field_y(self) -> StatusSignal[microtesla]:
        """
        The raw magnitude of the magnetic field measured by the Pigeon 2 in
        the Y direction. This is only valid after performing a magnetometer
        calibration.
        
          Minimum Value: -19660.8
          Maximum Value: 19660.2
          Default Value: 0
          Units: uT
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: RawMagneticFieldY Status Signal Object
        :rtype: StatusSignal[microtesla]
        """
        return self._common_lookup(SpnValue.PIGEON2_RAW_MAGNETIC_FIELDY.value, 0, None, "raw_magnetic_field_y", True, microtesla)
    
    def get_raw_magnetic_field_z(self) -> StatusSignal[microtesla]:
        """
        The raw magnitude of the magnetic field measured by the Pigeon 2 in
        the Z direction. This is only valid after performing a magnetometer
        calibration.
        
          Minimum Value: -19660.8
          Maximum Value: 19660.2
          Default Value: 0
          Units: uT
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: RawMagneticFieldZ Status Signal Object
        :rtype: StatusSignal[microtesla]
        """
        return self._common_lookup(SpnValue.PIGEON2_RAW_MAGNETIC_FIELDZ.value, 0, None, "raw_magnetic_field_z", True, microtesla)
    
    def get_is_pro_licensed(self) -> StatusSignal[bool]:
        """
        Whether the device is Phoenix Pro licensed.
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: IsProLicensed Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.VERSION_IS_PRO_LICENSED.value, 0, None, "is_pro_licensed", True, bool)
    
    def get_fault_hardware(self) -> StatusSignal[bool]:
        """
        Hardware fault occurred
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_Hardware Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_HARDWARE.value, 0, None, "fault_hardware", True, bool)
    
    def get_sticky_fault_hardware(self) -> StatusSignal[bool]:
        """
        Hardware fault occurred
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_Hardware Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_HARDWARE.value, 0, None, "sticky_fault_hardware", True, bool)
    
    def get_fault_undervoltage(self) -> StatusSignal[bool]:
        """
        Device supply voltage dropped to near brownout levels
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_Undervoltage Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_UNDERVOLTAGE.value, 0, None, "fault_undervoltage", True, bool)
    
    def get_sticky_fault_undervoltage(self) -> StatusSignal[bool]:
        """
        Device supply voltage dropped to near brownout levels
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_Undervoltage Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_UNDERVOLTAGE.value, 0, None, "sticky_fault_undervoltage", True, bool)
    
    def get_fault_boot_during_enable(self) -> StatusSignal[bool]:
        """
        Device boot while detecting the enable signal
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_BootDuringEnable Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_BOOT_DURING_ENABLE.value, 0, None, "fault_boot_during_enable", True, bool)
    
    def get_sticky_fault_boot_during_enable(self) -> StatusSignal[bool]:
        """
        Device boot while detecting the enable signal
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_BootDuringEnable Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_BOOT_DURING_ENABLE.value, 0, None, "sticky_fault_boot_during_enable", True, bool)
    
    def get_fault_unlicensed_feature_in_use(self) -> StatusSignal[bool]:
        """
        An unlicensed feature is in use, device may not behave as expected.
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_UnlicensedFeatureInUse Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_UNLICENSED_FEATURE_IN_USE.value, 0, None, "fault_unlicensed_feature_in_use", True, bool)
    
    def get_sticky_fault_unlicensed_feature_in_use(self) -> StatusSignal[bool]:
        """
        An unlicensed feature is in use, device may not behave as expected.
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_UnlicensedFeatureInUse Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_UNLICENSED_FEATURE_IN_USE.value, 0, None, "sticky_fault_unlicensed_feature_in_use", True, bool)
    
    def get_fault_bootup_accelerometer(self) -> StatusSignal[bool]:
        """
        Bootup checks failed: Accelerometer
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_BootupAccelerometer Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_BOOTUP_ACCEL.value, 0, None, "fault_bootup_accelerometer", True, bool)
    
    def get_sticky_fault_bootup_accelerometer(self) -> StatusSignal[bool]:
        """
        Bootup checks failed: Accelerometer
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_BootupAccelerometer Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_BOOTUP_ACCEL.value, 0, None, "sticky_fault_bootup_accelerometer", True, bool)
    
    def get_fault_bootup_gyroscope(self) -> StatusSignal[bool]:
        """
        Bootup checks failed: Gyroscope
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_BootupGyroscope Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_BOOTUP_GYROS.value, 0, None, "fault_bootup_gyroscope", True, bool)
    
    def get_sticky_fault_bootup_gyroscope(self) -> StatusSignal[bool]:
        """
        Bootup checks failed: Gyroscope
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_BootupGyroscope Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_BOOTUP_GYROS.value, 0, None, "sticky_fault_bootup_gyroscope", True, bool)
    
    def get_fault_bootup_magnetometer(self) -> StatusSignal[bool]:
        """
        Bootup checks failed: Magnetometer
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_BootupMagnetometer Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_BOOTUP_MAGNE.value, 0, None, "fault_bootup_magnetometer", True, bool)
    
    def get_sticky_fault_bootup_magnetometer(self) -> StatusSignal[bool]:
        """
        Bootup checks failed: Magnetometer
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_BootupMagnetometer Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_BOOTUP_MAGNE.value, 0, None, "sticky_fault_bootup_magnetometer", True, bool)
    
    def get_fault_boot_into_motion(self) -> StatusSignal[bool]:
        """
        Motion Detected during bootup.
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_BootIntoMotion Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_BOOT_INTO_MOTION.value, 0, None, "fault_boot_into_motion", True, bool)
    
    def get_sticky_fault_boot_into_motion(self) -> StatusSignal[bool]:
        """
        Motion Detected during bootup.
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_BootIntoMotion Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_BOOT_INTO_MOTION.value, 0, None, "sticky_fault_boot_into_motion", True, bool)
    
    def get_fault_data_acquired_late(self) -> StatusSignal[bool]:
        """
        Motion stack data acquisition was slower than expected
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_DataAcquiredLate Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_DATA_ACQUIRED_LATE.value, 0, None, "fault_data_acquired_late", True, bool)
    
    def get_sticky_fault_data_acquired_late(self) -> StatusSignal[bool]:
        """
        Motion stack data acquisition was slower than expected
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_DataAcquiredLate Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_DATA_ACQUIRED_LATE.value, 0, None, "sticky_fault_data_acquired_late", True, bool)
    
    def get_fault_loop_time_slow(self) -> StatusSignal[bool]:
        """
        Motion stack loop time was slower than expected.
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_LoopTimeSlow Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_LOOP_TIME_SLOW.value, 0, None, "fault_loop_time_slow", True, bool)
    
    def get_sticky_fault_loop_time_slow(self) -> StatusSignal[bool]:
        """
        Motion stack loop time was slower than expected.
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_LoopTimeSlow Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_LOOP_TIME_SLOW.value, 0, None, "sticky_fault_loop_time_slow", True, bool)
    
    def get_fault_saturated_magnetometer(self) -> StatusSignal[bool]:
        """
        Magnetometer values are saturated
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_SaturatedMagnetometer Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_SATURATED_MAGNE.value, 0, None, "fault_saturated_magnetometer", True, bool)
    
    def get_sticky_fault_saturated_magnetometer(self) -> StatusSignal[bool]:
        """
        Magnetometer values are saturated
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_SaturatedMagnetometer Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_SATURATED_MAGNE.value, 0, None, "sticky_fault_saturated_magnetometer", True, bool)
    
    def get_fault_saturated_accelerometer(self) -> StatusSignal[bool]:
        """
        Accelerometer values are saturated
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_SaturatedAccelerometer Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_SATURATED_ACCEL.value, 0, None, "fault_saturated_accelerometer", True, bool)
    
    def get_sticky_fault_saturated_accelerometer(self) -> StatusSignal[bool]:
        """
        Accelerometer values are saturated
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_SaturatedAccelerometer Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_SATURATED_ACCEL.value, 0, None, "sticky_fault_saturated_accelerometer", True, bool)
    
    def get_fault_saturated_gyroscope(self) -> StatusSignal[bool]:
        """
        Gyroscope values are saturated
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: Fault_SaturatedGyroscope Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.FAULT_PIGEON2_SATURATED_GYROS.value, 0, None, "fault_saturated_gyroscope", True, bool)
    
    def get_sticky_fault_saturated_gyroscope(self) -> StatusSignal[bool]:
        """
        Gyroscope values are saturated
        
          Default Value: False
        
        Default Rates:
          CAN: 4.0 Hz
        
        :returns: StickyFault_SaturatedGyroscope Status Signal Object
        :rtype: StatusSignal[bool]
        """
        return self._common_lookup(SpnValue.STICKY_FAULT_PIGEON2_SATURATED_GYROS.value, 0, None, "sticky_fault_saturated_gyroscope", True, bool)
    

    def set_control(self, request: SupportsSendRequest) -> StatusCode:
        """
        Control motor with generic control request object.

        If control request is not supported by device, this request
        will fail with StatusCode NotSupported

        :param request: Control object to request of the device
        :type request: SupportsSendRequest
        :return: StatusCode of the request
        :rtype: StatusCode
        """
        if isinstance(request, ()):
            return self._set_control_private(request)
        return StatusCode.NOT_SUPPORTED

    
    def set_yaw(self, new_value: degree, timeout_seconds: second = 0.050) -> StatusCode:
        """
        The yaw to set the Pigeon2 to right now.
        
        :param new_value: Value to set to. Units are in deg.
        :type new_value: degree
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.set_yaw(new_value, timeout_seconds)
    
    def clear_sticky_faults(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear the sticky faults in the device.
        
        This typically has no impact on the device functionality.  Instead, it
        just clears telemetry faults that are accessible via API and Tuner
        Self-Test.
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_faults(timeout_seconds)
    
    def clear_sticky_fault_hardware(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Hardware fault occurred
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_hardware(timeout_seconds)
    
    def clear_sticky_fault_undervoltage(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Device supply voltage dropped to near brownout
        levels
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_undervoltage(timeout_seconds)
    
    def clear_sticky_fault_boot_during_enable(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Device boot while detecting the enable signal
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_boot_during_enable(timeout_seconds)
    
    def clear_sticky_fault_bootup_accelerometer(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Bootup checks failed: Accelerometer
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_bootup_accelerometer(timeout_seconds)
    
    def clear_sticky_fault_bootup_gyroscope(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Bootup checks failed: Gyroscope
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_bootup_gyroscope(timeout_seconds)
    
    def clear_sticky_fault_bootup_magnetometer(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Bootup checks failed: Magnetometer
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_bootup_magnetometer(timeout_seconds)
    
    def clear_sticky_fault_boot_into_motion(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Motion Detected during bootup.
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_boot_into_motion(timeout_seconds)
    
    def clear_sticky_fault_data_acquired_late(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Motion stack data acquisition was slower than
        expected
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_data_acquired_late(timeout_seconds)
    
    def clear_sticky_fault_loop_time_slow(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Motion stack loop time was slower than expected.
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_loop_time_slow(timeout_seconds)
    
    def clear_sticky_fault_saturated_magnetometer(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Magnetometer values are saturated
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_saturated_magnetometer(timeout_seconds)
    
    def clear_sticky_fault_saturated_accelerometer(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Accelerometer values are saturated
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_saturated_accelerometer(timeout_seconds)
    
    def clear_sticky_fault_saturated_gyroscope(self, timeout_seconds: second = 0.050) -> StatusCode:
        """
        Clear sticky fault: Gyroscope values are saturated
        
        :param timeout_seconds: Maximum time to wait up to in seconds.
        :type timeout_seconds: second
        :returns: StatusCode of the set command
        :rtype: StatusCode
        """
    
        return self.configurator.clear_sticky_fault_saturated_gyroscope(timeout_seconds)

